//***************************************************************
//  This program includes just this file
//
//  Problem:
//  To imitate the problem Sleeping-Barber:
//      1. There are N wait-chairs for customers (represented by
//      WaitQueue::kMaxSize places in WaitQueue).
//      2. The barber can cut hair for just one customer
//      at a moment. If there's no customer waiting, the barber
//      will go to bed.
//      3. When a customer comes:
//        3.1. If the barber is sleeping => wake him up to serve
//        3.2. If the barber is working, check if there's any
//          wait-chair available, or else go away
//
//  Created:
//      Sep 24 2014
//  Last modified:
//      May 26 2016
//
//  Author:
//      Nguyen Duc Manh
//****************************************************************

// g++ ./SleepingBarber.cpp -o SleepingBarber -std=c++0x -pthread -lrt -O1

#include <stdarg.h>
#include <pthread.h>
#include <stdio.h>
#include <cstdlib>
#include <list>
#include <time.h>
#include <chrono>
#include <random>
#include <semaphore.h>
#include <string.h>

#define INFINITE -1

class Customer;

char gIntro[] = 
	"-------------------------------------------------\n" \
	"This program simulates the sleeping barber problem.\n" \
	"The barber shop has one barber, one barber chair, and ten chairs for waiting customers to sit on. " \
	"If there are no customers present, the barber sits down in the barber chair and fall asleep ." \
	"When a customer arrive, he has to wake up the sleeping barber. " \
	"If additional customers arrive while the barber is cutting a customer's hair, they " \
	"either sit down (if there are empty chairs) or leave the shop (if all chairs are full).\n" \
	"Let's see how the sleeping barber problem is simulated.\n" \
	"-------------------------------------------------"
;

// global variables
sem_t gHCout;
std::vector<Customer*> gCustomers; // referenced in menu,
pthread_mutex_t gGoAwayLock;
std::vector<Customer*> gGoAway; // have to lock

int timedWait(sem_t* pSem, unsigned miniSeconds)
{
    struct timespec absolute_time;
    int error;
    
	if (miniSeconds == INFINITE) {
		return sem_wait(pSem);
	}
	
    clock_gettime(CLOCK_REALTIME, &absolute_time);
    absolute_time.tv_sec += miniSeconds / 1000;
	absolute_time.tv_nsec += (miniSeconds % 1000) * 1000000;
 
    error = sem_timedwait(pSem, &absolute_time);
    return error;
}

int timedLock(pthread_mutex_t* pMutex, unsigned miniSeconds)
{
    struct timespec absolute_time;
    int error;
    
	if (miniSeconds == INFINITE) {
		return pthread_mutex_lock(pMutex);
	}

    clock_gettime(CLOCK_REALTIME, &absolute_time);
    absolute_time.tv_sec += miniSeconds / 1000;
	absolute_time.tv_nsec += (miniSeconds % 1000) * 1000000;

    error = pthread_mutex_timedlock(pMutex, &absolute_time);
    return error;
}

void mutexUnlock(pthread_mutex_t* pMutex)
{
	pthread_mutex_unlock(pMutex);
	usleep(1);
}

int semPost(sem_t* pSem)
{
	int error = sem_post(pSem);
	usleep(1);
	return error;
}

void concurrentPrintf(const char* format, ...)
{
	sem_wait(&gHCout);
	va_list args;
    va_start(args, format);

	vprintf(format, args);
	va_end(args);
	semPost(&gHCout);
}

class Customer {
private:
	#define ID_LEN 0x24
    char fId[ID_LEN];
	int delay;
	int cuttingTime;
	sem_t fHandle;
	pthread_t fHThread;
public:
    Customer() {
		if (sem_init(&fHandle, 0, 0) != 0) {
			perror("semaphore initilization");
			exit(1);
		}
		delay = 0;
		fId[0] = '\0';
	}
	virtual ~Customer() {
		sem_destroy(&fHandle);
	}
	
    Customer(char* id) {
		Customer();
		setId(id);
	}

    virtual void WaitForCuttingFinishing() {
		// announce:
		concurrentPrintf("\n<<Customer %s : waiting>>\n", fId);
		
		if (0 != sem_wait(&fHandle)) {
			perror("semaphore waiting");
			exit(1);
		}
	}

    virtual void Run();
    virtual char* GetId(){
        return fId;
    }
	
	virtual int getCuttingTime() {
		return cuttingTime;
	}
	
	virtual void setCuttingTime(int t) {
		cuttingTime = t;
	}

	virtual pthread_t* getHThread() {
		return &fHThread;
	}
	
	virtual int getDelay() {
		return delay;
	}
	
    virtual sem_t* GetHandle(){
        return &fHandle;
    }
	
	virtual void setId(char* id) {
		strncpy(fId, id, ID_LEN);
		fId[ID_LEN-1] = '\0';
	}
	
	virtual void setDelay(int delay) {
		this->delay = delay;
	}
	
	virtual void display() {
		printf("ID: %s\n", fId);
		printf("delay: %d miniseconds\n", delay);
		printf("cuttingTime: %d miniSeconds\n", cuttingTime);
	}
};
// Just one thread can Pop or Push WaitQueue
// at a moment.
class WaitQueue {
private:
    static const int kMaxSize = 6;
    static std::list<Customer*> fgHairQueue;
    static pthread_mutex_t fgLock;
public:
	friend void concurrentDisplayState();
	static void init() {
		if (pthread_mutex_init(&fgLock, NULL) != 0) {
			perror("mutex initilization");
			exit(1);
		}
	}

    //check status
    static bool IsEmpty() {
        return fgHairQueue.empty();
    }
    //check status
    static bool IsFull() {
        return ((int)fgHairQueue.size() == kMaxSize);
    }
    //Push customer into the end of WaitQueue
    static bool Push(Customer* pCustomer, int timeOut){
		int error = timedLock(&fgLock, timeOut);
		if (ETIMEDOUT == error) {
			return false;
		}
		if (0 != error) {
			perror("mutex waiting");
			exit(1);
		}
		bool pushed = false;
		if (!IsFull()) {
			fgHairQueue.push_back(pCustomer);
			pushed = true;
		}
		mutexUnlock(&fgLock);
		return pushed;
    }
    //save the Customer at the front of WaitQueue into customer and
    //remove that Customer from queue
    static Customer* Pop(int timeOut){
		int error = timedLock(&fgLock, timeOut);
		if (ETIMEDOUT == error) {
			return nullptr;
		}
		if (0 != error) {
			perror("mutex waiting");
			exit(1);
		}
        
		Customer* pCustomer = nullptr;
		if (!IsEmpty()) {
			pCustomer = fgHairQueue.front();
			fgHairQueue.pop_front();
		}
		mutexUnlock(&fgLock);
		return pCustomer;
    }
};

// static class declaration
std::list<Customer*> WaitQueue::fgHairQueue;
pthread_mutex_t WaitQueue::fgLock;

class Barber{
public:
    static pthread_mutex_t fgHWakeUpTurn;
    static sem_t fgHBed;
    //fgCustomerToCut points to the fsem_t that the barber will fisrt cut
    //right after he wakes up
    static Customer* fgCustomerToCut;
    static bool fgSleeping;
    static sem_t fgHReady;
	static pthread_mutex_t fgFinishedLock;
	static std::list<char*> fgFinished;
public:
	static void init() {
		if (pthread_mutex_init(&fgHWakeUpTurn, NULL) != 0) {
			perror("mutex initilization");
			exit(1);
		}
		if (sem_init(&fgHBed, 0, 0) != 0) {
			perror("semaphore initilization");
			exit(1);
		}
		if (sem_init(&fgHReady, 0, 0) != 0) {
			perror("semaphore initilization");
			exit(1);
		}
		fgCustomerToCut = nullptr;
		fgSleeping = false;
	}
    //used by the barber, pretending to sleep by calling pthread_mutex_lock
    static void GoToBed() {
        fgSleeping = true;
        // announce
		concurrentPrintf("\n[[Barber going to bed]]\n");

        //----------------------------------
		if (0 != sem_wait(&fgHBed)) {
			perror("semaphore waiting");
			exit(1);
		}
        // announce
		concurrentPrintf("\n[[Barber woken up]]\n");
		
        //----------------------------------
    }

private:
	static void addFinished(Customer* pCustomer) {
		#define BUF_LEN_ADD_FINISHED 0x44
		char* finished = (char*) operator new (BUF_LEN_ADD_FINISHED);
		//printf("addFinished of %s: %p\n", pCustomer->GetId(), finished);
		memcpy(finished, pCustomer->GetId(), ID_LEN);
		pthread_mutex_lock(&fgFinishedLock);
		fgFinished.push_back(finished);
		pthread_mutex_unlock(&fgFinishedLock);
	}
public:
	static void clearFinished() {
		pthread_mutex_lock(&fgFinishedLock);
		for(char* id: fgFinished) {
			free(id);
		}
		
		fgFinished.clear();
		
		pthread_mutex_unlock(&fgFinishedLock);
	}
    //pretend to cut by printing announcement to the screen
    static void CutHair() {
		concurrentPrintf("\n[[start cutting: customer %s]]\n",fgCustomerToCut->GetId());
		//sleep for an interval
        usleep(fgCustomerToCut->getCuttingTime()*1000);
		Customer* tmp = fgCustomerToCut;
        //announce to the screen
		concurrentPrintf("\n[[finish cutting: customer %s]]\n", tmp->GetId());
        //----------------------------------
    }
    //this function is used for imitating the actions of the barber
    static void* Run(void* parameter) {
        //announce the main thread that the barber is ready
        sem_post(&fgHReady);
        //----------------------------------
        while(true) {
            //check if there's any hairToCut
            if (nullptr != fgCustomerToCut) {
                //cut
                CutHair();
				addFinished(fgCustomerToCut);
				//announce the customer that their hair has been cut
				semPost(fgCustomerToCut->GetHandle());
				fgCustomerToCut = nullptr;
            }
            //take one customer out of the wait queue
            while (nullptr != (fgCustomerToCut = WaitQueue::Pop(INFINITE))) {
                //cut hair
                CutHair();
				addFinished(fgCustomerToCut);
				semPost(fgCustomerToCut->GetHandle());
				fgCustomerToCut = nullptr;
            }
            //if cannot take any customer out of the wait queue
            //sleep
            GoToBed();
        }
        return 0;
    }
	static Customer* getCurrentCustomer() {
		return fgCustomerToCut;
	}
    //check status (used by customers)
    static bool IsSleeping() {
        return fgSleeping;
    }
    //wake up the barber and save customer into fgCustomerToCut
    static bool WakeUpBarberGiveHair(Customer* pCustomer, unsigned timeOut) {
        bool wakeUpSucceeded = false;
		int error = timedLock(&fgHWakeUpTurn, timeOut);
		if (ETIMEDOUT == error) {
			return false;
		}
		if (0 != error) {
			perror("mutex waiting");
			exit(1);
		}
		
		if (IsSleeping()) {
			//attach fsem_t to fgCustomerToCut
			fgCustomerToCut = pCustomer;
			//wake up the barber
			fgSleeping = false;
			semPost(&fgHBed);
			wakeUpSucceeded = true;
		}
		mutexUnlock(&fgHWakeUpTurn);
		return wakeUpSucceeded;
    }
};

// static variables
pthread_mutex_t Barber::fgHWakeUpTurn;
sem_t Barber::fgHBed;
sem_t Barber::fgHReady;
Customer *Barber::fgCustomerToCut;
bool Barber::fgSleeping;
pthread_mutex_t Barber::fgFinishedLock;
std::list<char*> Barber::fgFinished;


#define N_CUSTOMERS  10
//function to be passed to CreateThread
void* RunCustomer(void* lpParameter) {
    Customer* C = (Customer*)lpParameter;
    C->Run();
    return 0;
}

void init()
{
	setvbuf(stdout, NULL, _IONBF, 0);
	setvbuf(stdin, NULL, _IONBF, 0);

	WaitQueue::init();
	Barber::init();
	
	if (pthread_mutex_init(&Barber::fgFinishedLock, NULL) != 0) {
		perror("mutex initilization");
		exit(1);
	}
	
	if (pthread_mutex_init(&gGoAwayLock, NULL) != 0) {
		perror("mutex initilization");
		exit(1);
	}
	
	if (sem_init(&gHCout, 0, 1) != 0) {
		perror("semaphore initialization");
		exit(1);
	}
}

void randomCustomers()
{
	char id[20];
	unsigned seed
            = std::chrono::system_clock::now().time_since_epoch().count();
    std::default_random_engine generator(seed);
    std::uniform_int_distribution<int> distributionA(0, 10000);
	std::uniform_int_distribution<int> distributionB(500, 2000);
	//create customer threads
	gCustomers.clear();
    for (int i = 0; i < N_CUSTOMERS; ++i) {
		sprintf(id, "%d", i);
        Customer* pCustomer = new Customer(id);
		pCustomer->setDelay(distributionA(generator));
		pCustomer->setCuttingTime(distributionB(generator));
        gCustomers.push_back(pCustomer);
    }
	printf("Created %d customers\n", N_CUSTOMERS);
}

void run()
{
	int rc;
	//create customer threads
	for (int i = 0; i < gCustomers.size(); ++i) {
		Customer* pCustomer = gCustomers[i];
        rc = pthread_create(pCustomer->getHThread(), NULL, RunCustomer, pCustomer);
        if (0 != rc) {
            fprintf(stderr, "error: pthread_create, rc: %d\n", rc);
			exit(1);
        }
    }
	/* block until all threads complete */
	for (int i = 0; i < gCustomers.size(); ++i) {
		Customer* pCustomer = gCustomers[i];
		pthread_join(*(pCustomer->getHThread()), NULL);
		//printf("deleted %s\n", pCustomer->GetId());
		delete pCustomer;
	}
	
	gCustomers.clear();
	gGoAway.clear();
	Barber::clearFinished();
	
	concurrentPrintf("\nFinished all customers.\n");
}


#define ADD 1
#define DELETE 2
#define CLEAR 3
#define RANDOM 4
#define SHOW 5
#define RUN 6
#define EXIT 7

void mDelete(unsigned index)
{
	if(index >= gCustomers.size()) {
		printf("Invalid index: %d\n", index);
	} else {
		Customer* pCustomer = *(gCustomers.begin()+index);
		gCustomers.erase(gCustomers.begin()+index);
		delete pCustomer;
		printf("Deleted\n");
	}
}

void* menu(void* p) // just have the barber, don't have to concurrent
{
	int choice;
	char id[ID_LEN];
	int t;
	while (1) {
		printf("Menu:\n");
		printf("%d: Add new customer\n", ADD);
		printf("%d: Delete a customer\n", DELETE);
		printf("%d: Delete all customers\n", CLEAR);
		printf("%d: Create random customers\n", RANDOM);
		printf("%d: Show customers\n", SHOW);
		printf("%d: Run\n", RUN);
		printf("%d: Exit\n", EXIT);
		printf("Your choice: ");
		
		scanf("%d", &choice);
		getchar();
		switch(choice) {
			case ADD:
			{
				if (gCustomers.size() >= N_CUSTOMERS) {
					printf("Full of customers. Cannot add anymore!\n");
				} else {
					printf("Input customer's ID: ");
					fgets(id, ID_LEN, stdin);
					id[strlen(id)-1] = '\0';
					
					Customer* pCustomer = new Customer(id);
					
					printf("Input customer's delay (in miniseconds): ");
					scanf("%d", &t);
					getchar();
					pCustomer->setDelay(t);
					
					printf("Input customer's cutting time (in miniseconds): ");
					scanf("%d", &t);
					getchar();
					pCustomer->setCuttingTime(t);
					//printf("Customer %s: %p", pCustomer->GetId(), pCustomer);
					gCustomers.push_back(pCustomer);
				}
			}
			break;
			case DELETE:
			{
				if (gCustomers.size() == 0) {
					printf("The customer list is empty!\n");
				} else {
					printf("Input customer's index: ");
					unsigned index;
					scanf("%u", &index);
					getchar();
					mDelete(index);
				}
			}
			break;
			case CLEAR:
			{
				for(Customer* pCustomer: gCustomers) {
					delete pCustomer;
				}
				gCustomers.clear();
				printf("Deleted all customers\n");
			}
			break;
			case SHOW:
			{
				printf("Customer list:\n");
				for(int i = 0; i < gCustomers.size(); ++i) {
					printf("------------------------------------\n");
					printf("Customer %d:\n", i);
					gCustomers[i]->display();
				}
			}
			break;
			case RUN:
			{
				run();
			}
			break;
			case RANDOM:
			{
				randomCustomers();
			}
			break;
			case EXIT:
			{
				exit(0);
			}
			break;
			default:
				printf("Invalid choice\n");
		}
	}
}

void concurrentDisplayState()
{
	#define BUF_LEN_DISPLAY 425
	char buf[BUF_LEN_DISPLAY];
	buf[0] = '\0';
	sem_wait(&gHCout);

	Customer* pCustomer = Barber::getCurrentCustomer();
	strcat(buf, "Curr: ");
	if (pCustomer != nullptr) {
		strcat(buf, "{");
		strcat(buf, pCustomer->GetId());
		strcat(buf, "}");
	}
	strcat(buf, "\n");
	// wait queue
	pthread_mutex_lock(&WaitQueue::fgLock);
	strcat(buf, "Wait Queue: ");
	for(Customer* pCustomer: WaitQueue::fgHairQueue) {
		strcat(buf, "[");
		strcat(buf, pCustomer->GetId());
		strcat(buf, "]");
	}
	strcat(buf, "\n");
	pthread_mutex_unlock(&WaitQueue::fgLock);
	// finished
	pthread_mutex_lock(&Barber::fgFinishedLock);
	if (Barber::fgFinished.size() != 0) {
		strcat(buf, "Finished: ");
		for(char* id: Barber::fgFinished) {
			strcat(buf, "[");
			strcat(buf, id);
			strcat(buf, "]");
		}
		strcat(buf, "\n");
	}
	pthread_mutex_unlock(&Barber::fgFinishedLock);
	// go away
	pthread_mutex_lock(&gGoAwayLock);
	if (gGoAway.size() != 0) {
		strcat(buf, "Go away: ");
		for(Customer* pCustomer: gGoAway) {
			//strcat(buf, "(");
			//strcat(buf, (char*)pCustomer);
			//strcat(buf, ")");
			strcat(buf, "[");
			strcat(buf, pCustomer->GetId());
			strcat(buf, "]");
		}
		strcat(buf, "\n");
	}
	pthread_mutex_unlock(&gGoAwayLock);
	strcat(buf,"------------------------\n");
	
	printf(buf);
	sem_post(&gHCout);
}

int main() {
	puts(gIntro);
	
	init();

	int rc;
    pthread_t hBarber;
	pthread_t hMenu;
	
    //create barber thread
	rc = pthread_create(&hBarber, NULL, Barber::Run, NULL);
    if (0 != rc) {
        fprintf(stderr, "error: pthread_create, rc: %d\n", rc);
		exit(1);
    }
    //wait for the barber to be ready
    sem_wait(&Barber::fgHReady);

	rc = pthread_create(&hMenu, NULL, menu, NULL);
    if (0 != rc) {
        fprintf(stderr, "error: pthread_create, rc: %d\n", rc);
		exit(1);
    }
	
	pthread_join(hBarber, NULL);
	pthread_join(hMenu, NULL);
	
	return 0;
}

void Customer::Run() {
    bool wakeUpResult, queueResult;
	usleep(delay*1000);
    while(true) {
        if (Barber::IsSleeping()) {
			concurrentPrintf("\n<<Customer %s: see Barber sleeping>>\n", fId);
            //wake up barber
			concurrentPrintf("\n<<Customer %s: try to wake up barber>>\n", fId);
            wakeUpResult = Barber::WakeUpBarberGiveHair(this, INFINITE);
            if (true == wakeUpResult) {
				concurrentPrintf("\n<<Customer %s: woken up barber successfully>>\n", fId);
				concurrentDisplayState();
                //wait for the cutting to finish
                WaitForCuttingFinishing();
                
				concurrentPrintf("\n<<Customer %s: go away after being cut>>\n", fId);
				concurrentDisplayState();
                //go away
                return;
            } else {
				concurrentPrintf("\n<<Customer %s: failed to wake up barber>>\n", fId);
                //get back to the
                //start position of the while loop
                continue;
            }
        } else {
			concurrentPrintf("\n<<Customer %s: see Barber not sleeping>>\n", fId);
            if (WaitQueue::IsFull()) {
				pthread_mutex_lock(&gGoAwayLock);
				gGoAway.push_back(this);
				pthread_mutex_unlock(&gGoAwayLock);
				
				concurrentPrintf("\n<<Customer %s: see WaitQueue full, go away>>\n", fId);
				concurrentDisplayState();
                //go away
                return;
            } else {
                //Push self to queue
				concurrentPrintf("\n<<Customer %s: see WaitQueue not full, try to get a chair to sit down>>\n", fId);
				concurrentDisplayState();

                queueResult = WaitQueue::Push(this, INFINITE);
                if (false == queueResult) {
					pthread_mutex_lock(&gGoAwayLock);
					gGoAway.push_back(this);
					pthread_mutex_unlock(&gGoAwayLock);
                    
					concurrentPrintf("\n<<Customer %s: failed to get a chair, go away>>\n", fId);
                    //I did not add my hair
                    // so I go away
                    return;
                } else {
					concurrentPrintf("\n<<Customer %s: got a chair successfully>>\n", fId);
					concurrentDisplayState();
                    //I'll wait for the cutting to finish
                    WaitForCuttingFinishing();
                    //and then go away
					concurrentPrintf("\n<<Customer %s: go away after being cut>>\n", fId);
					concurrentDisplayState();
                    return;
                }
            }
        }
    }
}
